var images=new Array();
var meshes=new Array();
var objects=new Array();

var imageProgress=0,imageCount=0;

var getObjectCollisionX=0,getObjectCollisionY=0,getObjectCollisionZ=0;

var deg2rad=Math.PI/180;
var cameraPositionX=0,cameraPositionY=0,cameraPositionZ=0,cameraAngleX=0,cameraAngleY=0,cameraAngleZ=0,cameraFOV=67.5,frustumAngle=360;

function getLookAt(angx,angy){
	return {x: Math.cos(angx)*Math.sin(angy), y: -Math.sin(angx), z: Math.cos(angy)*Math.cos(angx)};
}

function wrapvalue(a,b){
	if(!b)b=360;
	var a360=a/b;
	return (a360-Math.floor(a360))*b;
}

function RGBA(r,g,b,a){
	return new Pre3d.RGBA(r/255,g/255,b/255,a/255);
}

function RGB(r,g,b){
	return new Pre3d.RGBA(r/255,g/255,b/255,1.0);
}

function determ(a,b,c,d,e,f,g,h,i){
	return a*e*i+b*f*g+c*d*h-c*e*g-b*d*i-a*f*h;
}

function bind(scope, fn) {
    return function () {
        fn.apply(scope, arguments);
    };
}

function loadText(url){
	var req=new XMLHttpRequest();
	req.open('GET',url,false);
	req.send(null);
	if (req.status!=200)return '';
	return req.responseText;
}

function loadImage(i_file,i_num){
	imageCount++;

	var i_result=new Image();

	i_result.loaded=false;

	i_result.evalstring='';

	i_result.onload=function(){
		if(this.loaded==false){
			this.loaded=true;
			imageProgress++;
		}
		eval(this.evalstring);
	}

	i_result.src=i_file;

	if(i_result.complete==true&&i_result.loaded==false){
		i_result.loaded=true;
		image.progress++;
	}

	images[i_num]=i_result;
}

function imageExist(imgnum){
	var theimg=images[imgnum];
	if(theimg && theimg.loaded !== 'undefined' && (theimg.loaded==true||theimg.loaded==false)){
		return true;
	}else{
		return false;
	}
}

function raw2string(raw){
	var thestr=raw.replace(/[\s\t\n\r]+$/,'').replace(/[\n\r]+/g,';').replace(/[\s\t]+/g,',').replace(/\x2E0+,+/g,',').replace(/,+/g,',');
	if(raw.split(/[\n\r]+/)[0].indexOf('.')<0){
		/* If there is no dot on the first line, we assume it is a comment and therefore backtrack and get rid of the code crap we put there. */
		thestr=thestr.substring(thestr.indexOf(';')+1);
	}
	return (thestr+'a').replace(/\x2E0+a/g,'').replace(/a/g,'').replace(/,;/g,';');
}

function getMesh(meshstring){
	var s = new Pre3d.Shape();

	var thetris=meshstring.split(';');
	for(var y=0;y<thetris.length;y++){
		var thetri=thetris[y];
		if(thetri.length>0){
			var thetriarr=thetri.split(',');
			var v1={x: +thetriarr[0], y: +thetriarr[2], z: -thetriarr[1]};
			var v2={x: +thetriarr[3], y: +thetriarr[5], z: -thetriarr[4]};
			var v3={x: +thetriarr[6], y: +thetriarr[8], z: -thetriarr[7]};
			s.vertices.push(v1);
			s.vertices.push(v2);
			s.vertices.push(v3);
			s.quads.push(new Pre3d.QuadFace(y*3,y*3+1,y*3+2,null));
		}
	}

	Pre3d.ShapeUtils.rebuildMeta(s);

	return s;
}

function loadMesh(meshstring,meshnum){
	if(meshstring.substring(meshstring.length-3).toLowerCase()=='raw'){
		meshes[meshnum]=getMesh(raw2string(loadText(meshstring)));
	}else{
		meshes[meshnum]=getMesh(meshstring);
	}
}

function meshExist(meshnum){
	if(meshes[meshnum]){
		return true;
	}else{
		return false;
	}
}

function loadObject(meshstring,objnum){
	var themesh=new Pre3d.Shape();
	if(meshstring.substring(meshstring.length-3).toLowerCase()=='raw'){
		themesh=getMesh(raw2string(loadText(meshstring)));
	}else{
		themesh=getMesh(meshstring);
	}

	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: themesh,
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: false,
			meshtype: 0
		}]
	}
	resetObjectTransform(objnum);
}

function makeObjectBox(objnum,w,h,d){
	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: Pre3d.ShapeUtils.makeBox(w/2,h/2,d/2),
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: false,
			meshtype: 1
		}]
	}
	resetObjectTransform(objnum);
}

function makeObjectCube(objnum,s){
	makeObjectBox(objnum,s,s,s);
}

function makeObjectPlane(objnum,w,h){
	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: Pre3d.ShapeUtils.makePlane({x: -w/2, y: h/2, z: 0},{x: -w/2, y: -h/2, z: 0},{x: w/2, y: -h/2, z: 0},{x: w/2, y: h/2, z: 0}),
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: true,
			meshtype: 2
		}]
	}
	resetObjectTransform(objnum);
}

var makeObjectPlain=makeObjectPlane;

function makeObjectTriangle(objnum,x1,y1,z1,x2,y2,z2,x3,y3,z3){
	var theshape=new Pre3d.Shape();
	theshape.vertices=[{x: x1, y: y1, z: z1}, {x: x2, y: y2, z: z2}, {x: x3, y: y3, z: z3}];
	theshape.quads=[new Pre3d.QuadFace(0, 1, 2, null)];
	Pre3d.ShapeUtils.rebuildMeta(theshape);
	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: theshape,
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: true,
			meshtype: 0
		}]
	}
	resetObjectTransform(objnum);
}

function makeObjectSphere(objnum,s){
	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: Pre3d.ShapeUtils.makeSphere(s/2,10,5),
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: false,
			meshtype: 0
		}]
	}
	resetObjectTransform(objnum);
}

function makeObjectColorBox(objnum,w,h,d,rc,lc,uc,dc,fc,bc){

	var odraw1=false;
	if(rc.a>=1.0)odraw1=true;
	var odraw2=false;
	if(lc.a>=1.0)odraw2=true;
	var odraw3=false;
	if(uc.a>=1.0)odraw3=true;
	var odraw4=false;
	if(dc.a>=1.0)odraw4=true;
	var odraw5=false;
	if(fc.a>=1.0)odraw5=true;
	var odraw6=false;
	if(bc.a>=1.0)odraw6=true;

	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: Pre3d.ShapeUtils.makePlane({x: w/2, y: h/2, z: d/2},{x: w/2, y: -h/2, z: d/2},{x: w/2, y: -h/2, z: -d/2},{x: w/2, y: h/2, z: -d/2}),
			color: rc,
			overdraw: odraw1,
			backface: false,
			meshtype: 2
		},{
			mesh: Pre3d.ShapeUtils.makePlane({x: -w/2, y: h/2, z: -d/2},{x: -w/2, y: -h/2, z: -d/2},{x: -w/2, y: -h/2, z: d/2},{x: -w/2, y: h/2, z: d/2}),
			color: lc,
			overdraw: odraw2,
			backface: false,
			meshtype: 2
		},{
			mesh: Pre3d.ShapeUtils.makePlane({x: -w/2, y: -h/2, z: d/2},{x: -w/2, y: -h/2, z: -d/2},{x: w/2, y: -h/2, z: -d/2},{x: w/2, y: -h/2, z: d/2}),
			color: uc,
			overdraw: odraw3,
			backface: false,
			meshtype: 2
		},{
			mesh: Pre3d.ShapeUtils.makePlane({x: -w/2, y: h/2, z: -d/2},{x: -w/2, y: h/2, z: d/2},{x: w/2, y: h/2, z: d/2},{x: w/2, y: h/2, z: -d/2}),
			color: dc,
			overdraw: odraw4,
			backface: false,
			meshtype: 2
		},{
			mesh: Pre3d.ShapeUtils.makePlane({x: w/2, y: h/2, z: -d/2},{x: w/2, y: -h/2, z: -d/2},{x: -w/2, y: -h/2, z: -d/2},{x: -w/2, y: h/2, z: -d/2}),
			color: fc,
			overdraw: odraw5,
			backface: false,
			meshtype: 2
		},{
			mesh: Pre3d.ShapeUtils.makePlane({x: -w/2, y: h/2, z: d/2},{x: -w/2, y: -h/2, z: d/2},{x: w/2, y: -h/2, z: d/2},{x: w/2, y: h/2, z: d/2}),
			color: bc,
			overdraw: odraw6,
			backface: false,
			meshtype: 2
		}]
	}
	resetObjectTransform(objnum);
}

function makeObject(objnum,meshnum,imgnum){
	objects[objnum]={
		trans: new Pre3d.Transform(),
		meshes: [{
			mesh: meshes[meshnum],
			color: new Pre3d.RGBA(1.0,1.0,1.0,1.0),
			overdraw: true,
			backface: false,
			meshtype: 0
		}]
	}
	resetObjectTransform(objnum);
	if(imgnum>0)textureObject(objnum,imgnum);
}

function appendObjectMesh(objnum,meshnum,col){
	var odraw=false;
	if(col.a>=1.0)odraw=true;
	objects[objnum].meshes.push({
		mesh: meshes[meshnum],
		color: col,
		overdraw: odraw,
		backface: false,
		meshtype: 0
	});
}

function colorObject(objnum,col){
	var odraw=false;
	if(col.a>=1.0)odraw=true;
	var themeshes=objects[objnum].meshes;
	for(var i=0;i<themeshes.length;i++){
		var themesh=themeshes[i];
		themesh.color=col;
		themesh.overdraw=odraw;
		themesh.quad_callback=null;
		themesh.texture=null;
	}
}

function setObjectCull(objnum,docull){
	var bface=false;
	if(docull==false)bface=true;
	var themeshes=objects[objnum].meshes;
	for(var i=0;i<themeshes.length;i++){
		themeshes[i].backface=bface;
	}
}

function textureObject(objnum,imgnum){
	var tex=images[imgnum];

	if(tex.loaded==false){
		tex.evalstring+='if(objectExist('+objnum+'))textureObject('+objnum+','+imgnum+');';
		return;
	}

	var themesh=objects[objnum].meshes[0];

	var themeshes=objects[objnum].meshes;
	for(var i=0;i<themeshes.length;i++){
		var themesh=themeshes[i];

			var w = tex.width;
			var h = tex.height;

			var texinfo1 = new Pre3d.TextureInfo();
			texinfo1.image = tex;
			texinfo1.u0 = 0;
			texinfo1.v0 = 0;
			texinfo1.u1 = 0;
			texinfo1.v1 = h;
			texinfo1.u2 = w;
			texinfo1.v2 = h;

			var texinfo2 = new Pre3d.TextureInfo();
			texinfo2.image = tex;
			texinfo2.u0 = 0;
			texinfo2.v0 = 0;
			texinfo2.u1 = w;
			texinfo2.v1 = h;
			texinfo2.u2 = w;
			texinfo2.v2 = 0;

			var texinfo3 = new Pre3d.TextureInfo();
			texinfo3.image = tex;
			texinfo3.u0 = w;
			texinfo3.v0 = 0;
			texinfo3.u1 = 0;
			texinfo3.v1 = 0;
			texinfo3.u2 = 0;
			texinfo3.v2 = h;

			var texinfo4 = new Pre3d.TextureInfo();
			texinfo4.image = tex;
			texinfo4.u0 = w;
			texinfo4.v0 = 0;
			texinfo4.u1 = 0;
			texinfo4.v1 = h;
			texinfo4.u2 = w;
			texinfo4.v2 = h;

		if(themesh.meshtype>0){
			themesh.color=null;
			themesh.overdraw=true;

			Pre3d.ShapeUtils.triangulate(themesh.mesh);

			themesh.texture=null;

			if(themesh.meshtype==1){

				themesh.quad_callback=function(quad_face, quad_index, shape) {
					/* Each face is two triangles, the newly triangulated triangles last. */
					this.texture = quad_index < 6 ? texinfo3 : texinfo4;
					return false;
				}
			}else{
				themesh.quad_callback=function(quad_face, quad_index, shape) {
					/* Each face is two triangles, the newly triangulated triangles last. */
					this.texture = quad_index < 1 ? texinfo1 : texinfo2;
					return false;
				}
			}
		}else{
			themesh.color=null;
			themesh.overdraw=true;

			Pre3d.ShapeUtils.triangulate(themesh.mesh);

			themesh.quad_callback=null;

			themesh.texture=texinfo1;
		}
	}
}

function positionObject(objnum,x,y,z){
	var theobj=objects[objnum];
	theobj.positionx=x;
	theobj.positiony=y;
	theobj.positionz=z;
}

function rotateObject(objnum,x,y,z){
	var theobj=objects[objnum];
	theobj.anglex=x;
	theobj.angley=y;
	theobj.anglez=z;
}

function scaleObject(objnum,x,y,z){
	var theobj=objects[objnum];
	theobj.scalex=x;
	theobj.scaley=y;
	theobj.scalez=z;
}

function moveObject(objnum,amount){
	var theobj=objects[objnum];
	var newpos=getLookAt(theobj.anglex*deg2rad,theobj.angley*deg2rad);
	theobj.positionx+=newpos.x*amount;
	theobj.positiony+=newpos.y*amount;
	theobj.positionz+=newpos.z*amount;
}

function resetObjectTransform(objnum){
	var theobj=objects[objnum];
	theobj.positionx=0;
	theobj.positiony=0;
	theobj.positionz=0;
	theobj.anglex=0;
	theobj.angley=0;
	theobj.anglez=0;
	theobj.scalex=100;
	theobj.scaley=100;
	theobj.scalez=100;
	theobj.visible=true;
	deleteObjectCollisionBox(objnum);
}

function showObject(objnum){
	objects[objnum].visible=true;
}

function hideObject(objnum){
	objects[objnum].visible=false;
}

function deleteObject(objnum){
	objects[objnum]=null;
}

function objectPositionX(objnum){
	return objects[objnum].positionx;
}

function objectPositionY(objnum){
	return objects[objnum].positiony;
}

function objectPositionZ(objnum){
	return objects[objnum].positionz;
}

function objectAngleX(objnum){
	return objects[objnum].anglex;
}

function objectAngleY(objnum){
	return objects[objnum].angley;
}

function objectAngleZ(objnum){
	return objects[objnum].anglez;
}

function objectVisible(objnum){
	return objects[objnum].visible;
}

function objectExist(objnum){
	var theobj=objects[objnum];
	if(theobj && theobj.trans !== 'undefined' && theobj.trans){
		return true;
	}else{
		return false;
	}
}

function makeObjectCollisionBox(objnum,x1,y1,z1,x2,y2,z2){
	var theobj=objects[objnum];
	theobj.collisionon=true;
	theobj.x1=x1;
	theobj.y1=y1;
	theobj.z1=z1;
	theobj.x2=x2;
	theobj.y2=y2;
	theobj.z2=z2;
}

function deleteObjectCollisionBox(objnum){
	var theobj=objects[objnum];
	theobj.collisionon=false;
	theobj.x1=0;
	theobj.y1=0;
	theobj.z1=0;
	theobj.x2=0;
	theobj.y2=0;
	theobj.z2=0;
}

function setObjectCollisionOn(objnum){
	objects[objnum].collisionon=true;
}

function setObjectCollisionOff(objnum){
	objects[objnum].collisionon=false;
}

function checkCollision(ob1,ob2){
	var c1=objects[ob1];
	var c2=objects[ob2];
	if(c1.collisionon==true&&c2.collisionon==true&&
				c1.x2+c1.positionx>c2.x1+c2.positionx&&
				c1.y2+c1.positiony>c2.y1+c2.positiony&&
				c1.z2+c1.positionz>c2.z1+c2.positionz&&
				c1.x1+c1.positionx<c2.x2+c2.positionx&&
				c1.y1+c1.positiony<c2.y2+c2.positiony&&
				c1.z1+c1.positionz<c2.z2+c2.positionz){
		var dif1=c1.x2+c1.positionx-(c2.x1+c2.positionx),dif2=c1.x1+c1.positionx-(c2.x2+c2.positionx);
		if(dif1+dif2>0){
			var movex=dif2;
		}else{
			var movex=dif1;
		}
		dif1=c1.y2+c1.positiony-(c2.y1+c2.positiony);dif2=c1.y1+c1.positiony-(c2.y2+c2.positiony);
		if(dif1+dif2>0){
			var movey=dif2;
		}else{
			var movey=dif1;
		}
		dif1=c1.z2+c1.positionz-(c2.z1+c2.positionz);dif2=c1.z1+c1.positionz-(c2.z2+c2.positionz);
		if(dif1+dif2>0){
			var movez=dif2;
		}else{
			var movez=dif1;
		}
		var amx=Math.abs(movex),amy=Math.abs(movey),amz=Math.abs(movez);
		if(amx<=amy&&amx<=amz){
			if(amx>Math.abs(getObjectCollisionX))getObjectCollisionX=movex;
		}else if(amy<=amx&&amy<=amz){
			if(amy>Math.abs(getObjectCollisionY))getObjectCollisionY=movey;
		}else{
			if(amz>Math.abs(getObjectCollisionZ))getObjectCollisionZ=movez;
		}
		return true;
	}else{
		return false;
	}
}

function objectCollision(objnum1,objnum2){
	getObjectCollisionX=0;
	getObjectCollisionY=0;
	getObjectCollisionZ=0;
	var result=0;
	if(!objnum2||objnum2==0){
		var col1=objects[objnum1];
		for(var i=0;i<objects.length;i++){
			if(objnum1!=i){
				var col2=objects[i];
				if(col2 !== 'undefined' && col2){
					if(checkCollision(objnum1,i))result=i;
				}
			}
		}
	}else if(checkCollision(objnum1,objnum2)){
		result=1;
	}
	return result;
}

function cloneObject(objnum2,objnum1){
	var obj1=objects[objnum1];
	var obj2={
		trans: new Pre3d.Transform(),
		positionx: obj1.positionx,
		positiony: obj1.positiony,
		positionz: obj1.positionz,
		anglex: obj1.anglex,
		angley: obj1.angley,
		anglez: obj1.anglez,
		scalex: obj1.scalex,
		scaley: obj1.scaley,
		scalez: obj1.scalez,
		visible: obj1.visible,
		collisionon: obj1.collisionon,
		x1: obj1.x1,
		y1: obj1.y1,
		z1: obj1.z1,
		x2: obj1.x2,
		y2: obj1.y2,
		z2: obj1.z1,
		meshes: []
	}
	var obj1meshes=obj1.meshes;
	for(var i=0;i<obj1meshes.length;i++){
		var obj1mesh=obj1meshes[i];
		var obj2mesh={
			mesh: obj1mesh.mesh,
			color: obj1mesh.color,
			overdraw: obj1mesh.overdraw,
			backface: obj1mesh.backface,
			meshtype: obj1mesh.meshtype
		}
		if(obj1mesh.quad_callback)obj2mesh.quad_callback=obj1mesh.quad_callback;
		if(obj1mesh.texture)obj2mesh.texture=obj1mesh.texture;
		obj2.meshes[i]=obj2mesh;
	}
	objects[objnum2]=obj2;
}

function positionCamera(x,y,z){
	cameraPositionX=x;
	cameraPositionY=y;
	cameraPositionZ=z;
}

function rotateCamera(x,y,z){
	cameraAngleX=x;
	cameraAngleY=y;
	cameraAngleZ=z;
}

function moveCamera(amount){
	var newpos=getLookAt(cameraAngleX*deg2rad,cameraAngleY*deg2rad);
	cameraPositionX+=newpos.x*amount;
	cameraPositionY+=newpos.y*amount;
	cameraPositionZ+=newpos.z*amount;
}

function setCameraFOV(fov){
	cameraFOV=fov;
}

function sync(ren){
	ren.clearBackground();

	var camx=cameraPositionX,camy=cameraPositionY,camz=cameraPositionZ,camangx=cameraAngleX,camangy=cameraAngleY,maxang=frustumAngle/2,rad2deg=1/deg2rad;
	if(camangx>90&&camangx<=270){
		camangx=wrapvalue(camangx-180);
		camangy=wrapvalue(camangy+180);
	}

	var thetrans=ren.camera.transform;
	thetrans.reset();
	thetrans.translate(-cameraPositionX,-cameraPositionY,cameraPositionZ);
	thetrans.rotateY(cameraAngleY*deg2rad);
	thetrans.rotateX(cameraAngleX*deg2rad);
	thetrans.rotateZ(cameraAngleZ*deg2rad);

	ren.camera.focal_length=2/Math.tan(cameraFOV*deg2rad/2);

	for(var x=0;x<objects.length;x++){
		var theobj=objects[x];
		if(theobj && theobj.trans !== 'undefined' && theobj.trans && theobj.visible == true){
			var xdist=theobj.positionx-camx,ydist=theobj.positiony-camy,zdist=theobj.positionz-camz;

			var anglex=Math.atan2(Math.sqrt(xdist*xdist+zdist*zdist),ydist)*rad2deg+270-camangx;
			if(anglex>=180)anglex-=360;

			var angley=Math.atan2(xdist,zdist)*rad2deg+360-camangy;
			if(angley>=180)angley-=360;

			if(Math.sqrt(anglex*anglex+angley*angley)<=maxang){
				var thetrans=theobj.trans;

				thetrans.reset();
				if(theobj.scalex!=100||theobj.scaley!=100||theobj.scalez!=100)thetrans.scale(theobj.scalex/100,theobj.scaley/100,theobj.scalez/100);
				thetrans.rotateZ(-theobj.anglez*deg2rad);
				thetrans.rotateX(-theobj.anglex*deg2rad);
				thetrans.rotateY(-theobj.angley*deg2rad);
				thetrans.translate(theobj.positionx,theobj.positiony,-theobj.positionz);

				ren.transform=thetrans;
				var themeshes=theobj.meshes;
				for(var y=0;y<themeshes.length;y++){
					var themesh=themeshes[y];
					ren.draw_overdraw=themesh.overdraw;
					ren.draw_backfaces=themesh.backface;
					ren.fill_rgba=themesh.color;
					if(themesh.texture !== 'undefined' && themesh.texture){
						ren.texture=themesh.texture;
					}else{
						ren.texture=null;
					}
					if(themesh.quad_callback !== 'undefined' && themesh.quad_callback){
						ren.quad_callback=bind(ren,themesh.quad_callback);
					}else{
						ren.quad_callback=null;
					}
					ren.bufferShape(themesh.mesh);
				}
			}
		}
	}

	ren.drawBuffer();
	ren.emptyBuffer();
}
